"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const collection_utils_1 = require("collection-utils");
const Naming_1 = require("./Naming");
const Source_1 = require("./Source");
const Annotation_1 = require("./Annotation");
const Support_1 = require("./support/Support");
function lineIndentation(line) {
    const len = line.length;
    let indent = 0;
    for (let i = 0; i < len; i++) {
        const c = line.charAt(i);
        if (c === " ") {
            indent += 1;
        }
        else if (c === "\t") {
            indent = (indent / 4 + 1) * 4;
        }
        else {
            return { indent, text: line.substring(i) };
        }
    }
    return { indent: 0, text: null };
}
class Renderer {
    constructor(targetLanguage, renderContext) {
        this.targetLanguage = targetLanguage;
        this.emitNewline = () => {
            const nl = Source_1.newline();
            this.pushItem(nl);
            this._lastNewline = nl;
        };
        this.emitItem = (item) => {
            if (this._needBlankLine) {
                this.emitNewline();
                this._needBlankLine = false;
            }
            this.pushItem(item);
        };
        this.emitTable = (tableArray) => {
            if (tableArray.length === 0)
                return;
            const table = tableArray.map(r => r.map(sl => Source_1.sourcelikeToSource(sl)));
            this.emitItem({ kind: "table", table });
            this.emitNewline();
        };
        this.typeGraph = renderContext.typeGraph;
        this.leadingComments = renderContext.leadingComments;
        this._finishedFiles = new Map();
        this.startEmit();
    }
    startEmit() {
        this._currentEmitTarget = this._emitted = [];
        this._needBlankLine = false;
        this._preventBlankLine = true; // no blank lines at start of file
    }
    pushItem(item) {
        this._currentEmitTarget.push(item);
        this._preventBlankLine = false;
    }
    ensureBlankLine() {
        if (this._preventBlankLine)
            return;
        this._needBlankLine = true;
    }
    preventBlankLine() {
        this._needBlankLine = false;
        this._preventBlankLine = true;
    }
    emitLine(...lineParts) {
        if (lineParts.length === 1) {
            this.emitItem(lineParts[0]);
        }
        else if (lineParts.length > 1) {
            this.emitItem(lineParts);
        }
        this.emitNewline();
    }
    emitMultiline(linesString) {
        const lines = linesString.split("\n");
        const numLines = lines.length;
        if (numLines === 0)
            return;
        this.emitLine(lines[0]);
        let currentIndent = 0;
        for (let i = 1; i < numLines; i++) {
            const line = lines[i];
            const { indent, text } = lineIndentation(line);
            Support_1.assert(indent % 4 === 0, "Indentation is not a multiple of 4.");
            if (text !== null) {
                const newIndent = indent / 4;
                this.changeIndent(newIndent - currentIndent);
                currentIndent = newIndent;
                this.emitLine(text);
            }
            else {
                this.emitNewline();
            }
        }
        if (currentIndent !== 0) {
            this.changeIndent(-currentIndent);
        }
    }
    emitAnnotated(annotation, emitter) {
        const oldEmitTarget = this._currentEmitTarget;
        const emitTarget = [];
        this._currentEmitTarget = emitTarget;
        emitter();
        Support_1.assert(this._currentEmitTarget === emitTarget, "_currentEmitTarget not restored correctly");
        this._currentEmitTarget = oldEmitTarget;
        const source = Source_1.sourcelikeToSource(emitTarget);
        this.pushItem(Source_1.annotated(annotation, source));
    }
    emitIssue(message, emitter) {
        this.emitAnnotated(new Annotation_1.IssueAnnotationData(message), emitter);
    }
    changeIndent(offset) {
        if (this._lastNewline === undefined) {
            return Support_1.panic("Cannot change indent for the first line");
        }
        this._lastNewline.indentationChange += offset;
    }
    forEach(iterable, interposedBlankLines, leadingBlankLine, emitter) {
        const items = Array.from(iterable);
        let onFirst = true;
        for (const [i, [k, v]] of collection_utils_1.iterableEnumerate(items)) {
            if ((leadingBlankLine && onFirst) || (interposedBlankLines && !onFirst)) {
                this.ensureBlankLine();
            }
            const position = items.length === 1 ? "only" : onFirst ? "first" : i === items.length - 1 ? "last" : "middle";
            emitter(v, k, position);
            onFirst = false;
        }
    }
    forEachWithBlankLines(iterable, blankLineLocations, emitter) {
        const interposing = ["interposing", "leading-and-interposing"].indexOf(blankLineLocations) >= 0;
        const leading = ["leading", "leading-and-interposing"].indexOf(blankLineLocations) >= 0;
        this.forEach(iterable, interposing, leading, emitter);
    }
    indent(fn) {
        this.changeIndent(1);
        fn();
        this.changeIndent(-1);
    }
    assignNames() {
        return Naming_1.assignNames(this.setUpNaming());
    }
    finishFile(filename) {
        Support_1.assert(!this._finishedFiles.has(filename), `Tried to emit file ${filename} more than once`);
        const source = Source_1.sourcelikeToSource(this._emitted);
        this._finishedFiles.set(filename, source);
        this.startEmit();
    }
    render(givenOutputFilename) {
        this._names = this.assignNames();
        this.emitSource(givenOutputFilename);
        if (this._emitted.length > 0) {
            this.finishFile(givenOutputFilename);
        }
        return { sources: this._finishedFiles, names: this._names };
    }
    get names() {
        if (this._names === undefined) {
            return Support_1.panic("Names accessed before they were assigned");
        }
        return this._names;
    }
}
exports.Renderer = Renderer;
